( -*- forth -*- )
( host side of the target monitor )

( for the 16bit stc:
to save the current state, so it can be loaded with stc16/interactive.f
type 'saveforth image/monitor.f' )



host

( these are the target 1 byte monitor commands
for most things jsr would be enough, but we need a one
byte ack, so we throw in some more to lessen communication load )

: monitor-in   0x1 out ;  ( receive byte )
: monitor-out  0x2 out ;  ( send out byte )
: monitor-jsr  0x3 out ;  ( jump to subroutine )

: monitor-lda  0x8 out ;  ( set data pointer )
: monitor-ldp  0x9 out ;  ( set program pointer )
: monitor-ack  0xa out ;  ( send ack byte )
: monitor-bye  0xb out ;  ( exit interpreter )


( sync: after giving a command that has a long
execution time, it's best to insert an ack to not
overflow the pic's input buffer )
: sync     monitor-ack in drop ;

( util )
: merge    ( lo hi -- ) 8 << 0xff00 and swap 0xff and or ;
: 2in      ( -- x )     in in merge ;
: 2out     ( x -- )     lohi swap out out ;

( target access
these are target mode commands taking arguments
from the host stack. these words are similar to the
t words, only beginning with x )

( stack access )
: >target   ( x -- )   monitor-in out ;  lastxt <tpush> h!
: target>   ( -- x )   monitor-out in ;
: 2>target  ( x -- )   lohi swap >target >target ;

( code execution )
: xexecute  ( xt -- )  monitor-jsr lohi swap out out ; lastxt <texecute> h!


( protocol sync reset )
: [expect]_  forchar >r begin_ r> lit_ in_ dup_ emit_ =_ until_ nextchar ;
: [hexpect]_ forchar >r begin_ r> lit_ in_ dup_ .x_ cr_ =_ until_ nextchar ;

: ping    ping# [expect] PING ;
: hexping ping# [hexpect] PING ;

( code and data access through target a/p pointers )
: n@p+      ( n -- )    >target n@p+# ;
: n!p+      ( n -- )    >target n!p+# ;
: n@a+      ( n -- )    >target n@a+# ;
: n!a+      ( n -- )    >target n!a+# ;

: a!       ( x -- )   monitor-lda 2out ;
: p!       ( x -- )   monitor-ldp 2out ;
: @p+      ( -- x )   2 n@p+ in in merge ;
: @a+      ( -- x )   1 n@a+ in ;
: !a+      ( x -- )   1 n!a+ out ;

( direct data/code fetch/store )
: data@      ( a -- x ) a! @a+ 0xff and ;   ( byte access )
: code@      ( a -- x ) p! @p+ 0xffff and ; ( word access )
: data!      ( x a -- ) a! !a+ ;

( shortcuts for target mode )
: @# data@ ;
: !# data! ;


( target code buffer readout )
variable t
: t> t h@ t@ 2 t h+! ;
: pt! dup p! t h! ;



: block      ( b -- n ) 64 * ;
: line       ( l -- n ) 8 * ;


: .dline+  ( -- )    8 n@a+ 8 for in space .byte next cr ;
: .dblock+ ( -- )    8 for .dline+ next ;

: .dline   ( l -- )  line  a! .dline+ ;
: .dblock  ( b -- )  block a! .dblock+ ;


: .line+    ( -- )      8 n@p+ 4 for 2in space .word next cr ;
: .block+   ( -- )      8 for .line+ next ;

: .line     ( l -- )    line p!  .line+ ;
: .block    ( b -- )    block p! .block+ ;


: upload  ( src dst count -- )
    swap a! dup n!a+ for dup hc@ out 1 + next drop ;

: upload-tcode
    swap a! dup n!a+ for dup tc@ out 1 + next drop ;

: nextline?   ( -- ? )    1 line n@p+ -1 4 for t> 2in = and next ;
: nextblock?  ( -- ? )    -1 8 for nextline? and next ;

: line?      ( l -- ? )   line pt!  nextline? ;
: block?     ( b -- ? )   block pt! nextblock? ;

: nblocks    ( -- b )   there 1 - 1 block / 1 + ;
: nlines     ( -- l )   there 1 - 1 line  / 1 + ;
: blockalign ( -- )     nblocks block there! ;
: linealign  ( -- )     nlines  line there! ;


: synced?   ( -- ? )
    0 -1  ( n flag )
    nblocks for over block? and >r 1 + r> next nip ;

: (lineprog)   ( l -- )   line dup p! 8 n!p+ 4 for dup t@ 2out 2 + next drop fprog# sync ;
: (blockerase) ( b -- )   block p! ferase# sync ;
: (blockprog)  ( b -- )   8 * 8 for dup (lineprog) 1 + next drop ;

( safety measure to prevent overwriting boot code.
need special ram-based interpreter-less transfer for that.

0x200 bytes
0x40  lines
0x8   blocks )

: lineprog    0x40 max (lineprog) ;
: blockerase  0x8  max (blockerase) ;
: blockprog   0x8  max (blockprog) ;

( 0 al ! 1 ah ! 0 pl ! 0 ph !
ferase  8 for   8 for @a+ !p+ next   fprog  next )



: uemit [c] u emit cr ;
: bprogress 0xd emit dup block .x space ;

: spaces for space next ;

: clearprogress 0xd emit 20 spaces 0xd emit ;

: blockupdate ( b -- )
    dup block? not if
	bprogress uemit
	dup blockerase blockprog
    else drop then clearprogress
;



( scratch block @ top of memory. )
511 constant scratchblock
: scratch   scratchblock block <texecute> h@ execute ;

: {#
    there scratchblock block there! ] ;

: upload-scratch 
    scratchblock blockupdate
    scratch sync ;

: },
    exit, there! [,
    upload-scratch ;

: codesync
    begin dup nblocks < while
	    bprogress dup blockupdate 1 +
    repeat drop ;

( update will upload application code, and leaves the monitor intact.
we use the target buffer's current compilation point as the start of
the updatable region. )

: update [ <t'> --user-- 64 / ]L codesync ;


( automatic incremental code update mode:
this does no block erase, but assumes the memory is cleared [all bits set],
so incremental writes can go by line.

verification:
each programmed line is checked. if check fails, an exclamation
point is output. normally, this should not occur since we scan
for a free line when entering incremental mode. )

variable last-line

: freeline? ( l -- ? )
    line p! 8 n@p+
    0xff 8 for in and next
    0xff = ;

: freeblock? ( b -- ? )
    block p! 64 n@p+
    0xff 64 for in and next
    0xff = ;

: i-align
    nlines begin
	dup freeline? not while 1 + repeat
    line there! ;

: i-mark
    linealign
    nlines last-line h@ = if exit then
    [saveforth] image/current.f
    nlines last-line h! ;

: lprogress 0xd emit dup line .x space ;
: i-sync
    last-line h@ begin
	dup nlines < while
	    lprogress [c] w emit
	    dup lineprog
	    dup line? not
	    if [c] ! emit cr then 1 +
    repeat drop clearprogress ;

( enter incremental mode. this aligns the
compilation point to 8 bytes, marks the last line,
and installs an execute vector that uploads non-synced
code before executing a target word )
: incremental
    i-align i-mark
    vector <texecute> i-sync i-mark xexecute ;

: erase-until
    >r 8 begin
	dup r < while
	    bprogress [c] e emit
	    dup blockerase 1 +
    repeat r> 2drop clearprogress ;


( this erases all application blocks.
need to empty and reload the interactive mode after this )
: erase-application
    nblocks erase-until recompile ;


( util )
: .fsr    0x3e .dblock .dblock+ ;
: .stack  data@ >r dup a! begin dup r <= while @a+ . 1 + repeat rdrop ;
: .ds     0x81 FSR0L .stack ;
: .xs     0x90 FSR1L .stack ;
: usage#  there 2 / ;

: .scratch scratchblock block dup 64 + .dasm ;

: p TBLPTRL data@ TBLPTRH data@ merge ;



( some remarks.

the scratch buffer is ok for some things, but has got a problem
with not resetting there on error. on the other hand it might be
wise to make the context saving automatic for incremental mode.

so we need:
* automatic saving of forth image. [incremental?]
* option to erase application. 

)

