
/* PICMICRO FORTH COMPILER / INTERPRETER

   2 machines
     host running compiler/interpreter
     target running target forth

   2 dictionaries (both stored on host side)
     host:    host code (compiler) and target macros (target + language)
     target:  compiled target code

   both are actually one dictionary so words can override macros and
   vice versa. this is equivalent to the 'immediate/postpone' approach, 
   only implemented using name mangling.

   the host execution model is an indirect threaded forth.

   the target execution model is defined in forth.

*/


#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#define TRUE -1 /* all one */
#define FALSE 0 /* all zero */

struct item_;
typedef struct item_ item_t;

struct item_
{
    item_t *prev;   // previous dict item
    char *name;
    int bytes;
    void *address;  // code address
    int target;     // address points to HOST or TARGET code
};


#define HOST   0  // compiler (host) dictionary
#define TARGET 1  // native code running on target
#define MACRO  2  // only parsing mode

typedef void (*prim_t)();    // address of primitive code
typedef struct stream_ stream_t;
typedef int word_t; // host word type (native signed int)


// the data cell has the size of the host pointer size. (32 bit)
// the target dicts have a smaller size (8 for ram, 16 for native code)


/* error messages */
void error_abort(char *msg);
void error_undef(char *name){
    char tmp[256];
    sprintf(tmp, "undefined word '%s'", name);
    error_abort(tmp);
}
void error_file(char *name) {
    char tmp[256];
    sprintf(tmp, "can't open file %s", name);
    error_abort(tmp);
}


/* stacks */
typedef struct {
    int size;
    void **data;
    void **top;
} safestack_t;

safestack_t *stack_new(int size)
{
    safestack_t *stack = calloc(1, sizeof(*stack));
    stack->size = size;
    stack->data = calloc(size, sizeof(void*));
    stack->top = stack->data + size;
    return stack;
}


void *safestack_top(safestack_t *s){return *(s->top);}

int stack_get_rsp(safestack_t *s){return s->top - s->data;}
void stack_set_rsp(safestack_t *s, int top){
    if ((top < 0) || (top > s->size))
	error_abort("stack pointer out of bound");
    else{
	s->top = s->data + top;
    }
}

void stack_push(safestack_t *s, void *thing){
    if (s->top <= s->data){
	error_abort("stack overflow");
    }
    else {
	s->top--;
	*(s->top) = thing;
    }

}
void *stack_pop(safestack_t *s){
    if (s->top >= (s->data + s->size)){
	error_abort("stack underflow");
    }
    else {
	void *x = *(s->top);
	s->top++;
	return x;
    }
}
void stack_clear(safestack_t *s){
    while (s->top < (s->data + s->size)) stack_pop(s);
}
void output_printf(char *fmt, ...);

void stack_print(char *fmt, safestack_t *s){
    void **i = s->data + s->size;
    output_printf("<%d> ", s->data + s->size - s->top);
    while (--i >= s->top) output_printf(fmt, *i);
    output_printf("\n");
}

/* global stacks */
safestack_t *ds, *rs, *is, *os;
void push_rs(prim_t *p){stack_push(rs, p);}
void push_ds(word_t i) {stack_push(ds, (void *)i);}
void push_is(stream_t*s)  {stack_push(is, (void *)s);}
void push_os(stream_t *s)     {stack_push(os, (void *)s);}
prim_t*   pop_rs() {return (prim_t *)stack_pop(rs);}
word_t    pop_ds() {return (int)stack_pop(ds);}
stream_t* pop_is() {return (stream_t *)stack_pop(is);}
stream_t* pop_os() {return (stream_t *)stack_pop(os);}
prim_t*   top_rs() {return *((prim_t **)(rs->top));}
word_t    top_ds() {return *((int *)(ds->top));}
stream_t* top_is() {return *((stream_t **)(is->top));}
stream_t* top_os() {return *((stream_t **)(os->top));}


/* global state */
int interactive = 0;
char word[64];           // last read input word
int current_target = 0;
int compile_mode = 0;    // compile mode
int port_fd = -1;        // serial port fd

/* host threaded code buffer */
#define HOST_SIZE 0x80000
prim_t host_buffer[HOST_SIZE];

/* current & marked pointers */
prim_t *host_top = host_buffer;
prim_t *last_xt  = 0;
item_t *dict_top = 0;

prim_t *mark_host_top = host_buffer;
item_t *mark_dict_top = 0;

void ***interpreter_start;  // interpreter ip (vector)
item_t *iword;              // dictionary traversal state


/* allocation */
void *allot(int bytes){
    char *c = (char *)host_top;
    void *v = c;
    c += bytes;
    host_top = (prim_t *)c;
    return v;
}
void align(void){
    int i = (int)host_top;
    i = (((i-1)/sizeof(int))+1)*sizeof(int);
    host_top = (prim_t *)i;
}


static inline int equal(char *a, char *b){
    while(1){
	if (a[0] != b[0]) return 0;
	if (!a[0]) return -1;
	a++;
	b++;
    }
}


/* * * * * * * * * * * *
*      STREAM IO       *
* * * * * * * * * * * */ 

typedef int (*nextchar_t)(stream_t *s);
struct stream_ {
    char *name;
    FILE *f;
    int line;
    int column;
    nextchar_t nextchar;
};


int stream_nextchar(stream_t *s){return fgetc(s->f);}

stream_t *stream_create(char *name){
    stream_t *s = calloc(sizeof(*s),1);
    s->name = malloc(strlen(name)+1);
    s->nextchar = stream_nextchar;
    strcpy(s->name, name);
    return s;
}

stream_t *stream_open_file(char *name, char *mode){
    stream_t *s = stream_create(name);
    if (s->f = fopen(name, mode)){
	return s;
    }
    else {
	free(s);
	return 0;
    }
}
stream_t *stream_open_fd(int fd, char *mode){
    stream_t *s = stream_create("-");
    if (s->f = fdopen(fd, mode)){
	return s;
    }
    else {
	free(s);
	return 0;
    }
}

void stream_close(stream_t *s){
    fclose(s->f);
    free(s->name);
    free(s);
}



/* * * * * * * * * * * * * *
*         OUTPUT           *
* * * * * * * * * * * * * */

void drop_os(){ 
    stream_close(pop_os());
    if (!top_os()) exit(0); // exit at end of output
}

void push_os_file(char *filename){
    stream_t *s;
    if (equal(filename, "-")){s = stream_open_fd(1, "w");}
    else {s = stream_open_file(filename, "w");}
    if (!s){error_file(filename);}
    else push_os(s);
}
void output_printf(char *fmt, ...){
    va_list ap;
    va_start(ap, fmt);
    vfprintf(top_os()->f, fmt, ap);
    va_end(ap);
    fflush(top_os()->f);
}

/* * * * * * * * * * * * * *
*          INPUT           *
* * * * * * * * * * * * * */

int wordis(char *w){return equal(word, w);}

void drop_is(){
    stream_close(pop_is());
    if (!top_is()) exit(0); // exit at end of input
    //if (equal("-", top_is()->name)) output_printf("done\n");
}
stream_t *stream_open_readline(char *filename);
void push_is_file(char *filename){
    stream_t *s;
    //fprintf(stderr, "* %s\n", filename);
    if (equal(filename, "-")){
	interactive= 1;
	if (!(s = stream_open_readline(filename)))
	    s = stream_open_fd(0, "r");
    }
    else {s = stream_open_file(filename, "r");}
    
    if (!s){error_file(filename);}
    else{push_is(s);}
}
char input_getc(void){
    char c;
    while ((c = top_is()->nextchar(top_is())) == EOF){
	drop_is();
    }
    // keep track of location of last character
    if (c == '\n'){
	top_is()->column = 0;
	top_is()->line++;
    }
    else {
	top_is()->column++;
    }
    return c;
}


void nextword(){
    char *pc = word;
  again:
    pc[0] = ' ';

    // skip whitespace
    do {pc[0] = input_getc();} while (isspace(pc[0]));
    
    // skip unix style comments
    // these are considered to be 'not there'
    // forth style comments are handled in forth
    // DISABLED: not really necessary. solve #! with a hack
    //           
    // if (pc[0] == '#'){
    //     do {pc[0] = input_getc();} while(pc[0] != '\n');
    //	   goto again;
    // }

    do {*++pc = input_getc();} while (!isspace(pc[0]));

    pc[0] = 0;

}




/* * * * * * * * * * * * * *
*       READLINE           *
* * * * * * * * * * * * * */

#define HAVE_READLINE
#ifdef HAVE_READLINE
#include <readline/readline.h>
#include <readline/history.h>

struct wordlist_;
typedef struct wordlist_ wordlist_t;

struct wordlist_ {
    wordlist_t *next;
    char *word;
};

wordlist_t *wordlist = 0;

void wordlist_push(char *c, int l){
    wordlist_t *w = malloc(sizeof(*w));
    w->next = wordlist;
    w->word = malloc(l+1);
    strncpy(w->word, c, l);
    w->word[l] = 0;
    // fprintf(stderr, "PUSH: (%s) %s\n", c, w->word);
    wordlist = w;
}
char *wordlist_pop(void){
    wordlist_t *w = wordlist;
    if (!w) return 0;
    char *c = w->word;
    wordlist = w->next;
    free(w);
    return c;
}

// this fills the wordlist
void words(void);
void macros(void);
void completions(void);

// generator for (current) dictionary words
char *command_generator (char *text, int state){
    static int len = 0;
    char *c;

    if (!state)	{
	// this just gives all the words, without their attribute suffix,
	// because it's quite impossible right now to list all the possible words
	// when editing a line, since there's no way to know which mode we're in.
	completions();
	len = strlen(text);
    }

  next:
    c = wordlist_pop();
    if (!c) return 0;
    if (strncmp(text, c, len)) { free(c); goto next; }
    return c;
}

char *histfile = ".badnop-history";

void setup_readline(void){
    using_history();
    if(read_history(histfile)){
	fprintf(stderr, "Creating new history file %s\n", histfile);
    }
    rl_completion_entry_function =  (rl_compentry_func_t *)command_generator;
    rl_basic_word_break_characters = " \t\n";
}

// generator for next character
int readline_nextchar(stream_t *s){
    static char *line = 0;
    static char *next = 0;
    char *prompt[] = {" ok (H)\n", " ok (T)\n", " ok (M)\n", " ok (L)\n"};
    if (!line){
	next = line = readline(" ok\n" /* prompt[current_target] */ );
	if (!line){
	    write_history(histfile);
	    output_printf("\n");
	    return EOF;
	}
	add_history(line);
    }
    if (!next[0]){
	free(line);
	line = 0;
	return '\n';
    }
    return *next++;
}

stream_t *stream_open_readline(char *name){
    stream_t *s = stream_open_fd(0, "r");
    s->nextchar = readline_nextchar;
    return s;
}


#else
stream_t *stream_open_readline(char *name){return 0;}
void setup_readline(){}
#endif



/* * * * * * * * * * *
*       UNIX         *
* * * * * * * * * * */

#ifdef __unix__
#include <unistd.h>
#include <termios.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <setjmp.h>

sigjmp_buf jump_abort;
static void sig(int n)
{
    char *msg = (char *)strsignal(n);
    error_abort(msg);
}

static void install_signal_handler(int signal, void (*handler)(int))
{
    struct sigaction sa;
    sa.sa_flags = 0;
    sa.sa_handler = handler;
    memset(&sa.sa_mask, 0, sizeof(sa.sa_mask));
    //sa.sa_restorer = 0;
    if (sigaction(signal, &sa, 0) < 0)
        output_printf("Can't set signal %d handler.", signal);

}



void setup_unix(void){
    install_signal_handler(SIGINT, sig);
    install_signal_handler(SIGSEGV, sig);
    install_signal_handler(SIGFPE, sig);
}

#else
// fake stubs
typedef int sigjmp_buf;
void siglongjmp (sigjmp_buf, int) { exit(1); }
int sigsetjmp (sigjmp_buf, int) { return 0; }
void setup_unix(void) { }
#endif


/* * * * * * * * * * *
*       ERROR        *
* * * * * * * * * * */

void error_where(void){
    stream_t *s = top_is();
    if (s) // print formatted error message
	fprintf(stderr, "%s:%d input: '%s' ", 
		s->name, s->line + (s->column ? 1 : 0), word);
}

void error_abort(char *msg){ 
    error_where();
    fprintf(stderr, "error: %s\n", msg);
    if (!interactive){exit(1);}
    siglongjmp(jump_abort, -1);
}

void warning(char *msg){ 
    error_where();
    fprintf(stderr, "warning: %s\n", msg);
}



/* * * * * * * * * * * * * *
*       DICTIONARY         *
* * * * * * * * * * * * * */


void new_item(int target, char *name)
{
    item_t *i = (item_t *)allot(sizeof(item_t));
    i->prev = dict_top;
    dict_top = i;
    //fprintf(stderr, "%s %x\n", name, dict_top);
    i->target = target;
    i->address = 0;
    i->bytes = 0;
    int bytes = strlen(name) + 2;
    i->name = (char *)allot(bytes); align();
    strcpy(i->name, name);
    i->name[bytes-1] = 0;
}



static inline void lookup_init(void){
    iword = dict_top;
}

static inline void lookup_continue(void){
    iword = iword->prev;
}
  
/*
  2 step matching: 
  - if input or dict item has a zero, and the other has a next zero, we have a preliminary match
  - depending on the kind of match, we can route to the appropriate handler.

 returns match signature:
 -1 : not found
*/

#define MATCH(inattrib, dictattrib, target) \
((inattrib & 0xff) | ((dictattrib & 0xff) << 8) | (target << 16))

int lookup_tick(char *name){
    int i;
    char *name2;

  check:
    if (iword == 0) return -1; // NOT FOUND
    name2 = iword->name;
    i = -1;
    //fprintf(stderr, "check\t%s\t%s\n", name, name2);

  nextchar:
    i++;
    if (name[i] && name2[i]){
	// both are nonzero: need to be equal
	if (name[i] == name2[i]) goto nextchar; 
	else goto nextword;
    }
    else {
	// word end found in one of the words. if the other
	// word is one char longer, it counts as a match, and the
	// extra char is the attribute.

	if (name[i] == 0) { if (name2[i+1] != 0) goto nextword; }
	else { if (name[i+1] != 0) goto nextword; }

	//fprintf(stderr, "match\t%s\t%s\n", name, name2);

	// match: compute signature, and pass control
	int t = iword->target;
	return MATCH(name[i], name2[i], t);
    }

  nextword:
    lookup_continue();
    goto check;

}
    

item_t *rlookup(void *x){
    item_t *i = dict_top;
  next:
    if (i == 0) return 0; // not found
    if (i->address == x) return i;
    i = i->prev;
    goto next;
}

  
/* dictionary listing */

void words_add(char attrib, int target){
    item_t *i = dict_top;
    int length;
  check:
    // fprintf(stderr, "%x\n", i);
    if (i == 0) return; // done
    if (i->target != target) goto next;
    length = strlen(i->name);
    if (attrib){
	if (length < 2) goto next;
	if (i->name[length-1] != attrib) goto next;
	wordlist_push(i->name, length-1);
	goto next;
    }
    wordlist_push(i->name, length);
  next:
    i = i->prev;
    goto check;
}


/* just dump entire dictionary, without the attribute suffixes.
this is easiest to use for command completion, and makes the list
a bit smaller. */ 
void completions(void){
    item_t *i = dict_top;
    int length;
  check:
    if (i == 0) return; // done
    length = strlen(i->name);
    if (length >= 2){
	switch (i->name[length-1]){
	case ',':
	case '_':
	case '%':
	case '#':
	    length--;
	    break;
	default:
	    break;
	}
    }
    wordlist_push(i->name, length);
    i = i->prev;
    goto check;    
}


void macros(void){
    switch(current_target){
    case HOST: 
	words_add('_', HOST);  // only host macros
	break;
    case TARGET:
    case MACRO:
	words_add(',', HOST);  // only target macros
	break;
    }
}

void words(void){
    switch(current_target){

	// basic modes
    case HOST: 
	words_add(0, HOST);  // literal host words
	break;
    case TARGET:
    case MACRO:
	words_add(0, TARGET); // literal target words
	break;

    }
}

// dump dictionary to current output

void print_wordlist(void)
{
    char *c;
    while (c = wordlist_pop()){
	output_printf("%s ", c);
	free(c);
    }
    output_printf("\n");
}


// setup input and output

#define FLAG(f) flag(*argv, f)

static char *flag(char *a, char *f){
    int n = strlen(f);
    if (!strncmp(a, f, n)){
	// fprintf(stderr, "flag: %s\n", a);
	return a + n;
    }
    return 0;
}

int setup_io(int argc, char **argv){
    char *def = "-";
    argc--;
    argv++;

    /* read flags */
    while(argc){
	char *val;
	if (val = FLAG("--log=")) { histfile = val; }
	else if (val = FLAG("--port=")) { 
	    port_fd = open(val, O_RDWR);
	    if (-1 == port_fd){
		fprintf(stderr, "can't open port %s\n", val);
	    }
	}
	else if (FLAG("--")) { fprintf(stderr, "unknown flag %s\n", *argv); exit(1); }
	else break;
	argc--;
	argv++;
    }
    
    setup_readline();

    /* connect io */
    push_os(0);                  // fallback output = NULL (terminate program)
    push_os_file("-");           // main output = stdout

    push_is(0);                  
    if (argc == 0){              // fallback input = NULL (terminate program)
	argc++;
	argv = &def;
    }
    while (argc--){
	push_is_file(argv[argc]);
    }
}



/* * * * * * * * * * * *
*  OUTER INTERPRETER   *
* * * * * * * * * * * */

/* 
   indirect threaded code interpreter

   threaded code is built of pointers to definition code field,
   immediately followed by data field(s)   [ CF | DF | DF | ... ]

   the code field points to the primitive to execute for this word (goto label)
   the w register contains the last CF pointer

   a code pointer (i.e. CF) is of type void* (goto label)
   a thread pointer is of type void** (points to a word (CF))
   an instruction pointer is of type void*** (points to a thread pointer)
   this means that code is of type void
*/


#define PUSH_RS(x) push_rs((prim_t *)(x))
#define POP_RS(x)  ((void ***)(pop_rs()))

#define PUSH_DS(x) push_ds((int)(x))
#define POP_DS(x)  ((int)(pop_ds()))

// parse a number
inline int getnumber(char *s, int *i){
    char *c;
    *i = strtoll(s, &c, 0);
    if (c[0]) {*i = 0; return FALSE;}
    else {return TRUE;}
}

/* xt cache */
void *xt_exit;
void *xt_mark;
void *xt_accept;
void *xt_interpret;
void *xt_quit;
void *xt_jmp;
void *xt_lit;
void *xt_bar;

void *xt_vcompilecomma;
void *xt_vlitcomma;
void *xt_vtexecute;
void *xt_vtpush;



/* NOTE:
   this is a rewrite of version 0.2 to a 'proper' indirect threaded
   forth, i.e. abandoning the c stack, to give full power
   to the forth return stack and related abuse.

   ( to assure you that this really is 'structured programming' )
*/


// forth execute environment is contained in this stack frame.
// the c stack is only used for calling c functions.
// all primitives here end on NEXT or PASS (goto other prim)
// if they don't, they obvously fall through
int main (int argc, char **argv)
{
    void ***ip;  // instruction pointer: points to next threaded instruction
    void **w;    // working register: points to current (executing) CF
    int i;       // fast temp var

    /* setup stacks */
    ds = stack_new(32);
    rs = stack_new(32);
    os = stack_new(32);
    is = stack_new(32);

    /* setup connection to the outside world */
    setup_io(argc, argv);
    setup_unix();

    /* convenience macros */
#define NEXT    goto next
#define OP2(op) {int i = pop_ds(); int j = pop_ds(); push_ds(j op i); NEXT;}
#define CMP(op) {int i = pop_ds(); int j = pop_ds(); push_ds((j op i) ? -1 : 0); NEXT;}
#define ADDR(addr)          {dict_top->address = last_xt = addr;}
#define PRIM(name, cname)   {WORD(name); ADDR(host_top); COMPILE(&&cname);}
#define VARIABLE(name, val) {WORD(name); ADDR(host_top); COMPILE(&&dovar); COMPILE(val);}
#define COMPILE(x)   {*host_top++ = (prim_t)(x);} // compile a word
#define WORD(name)   {new_item(HOST, name);}      // create host word
#define TARGETDICT(name, addr){new_item(TARGET, name); dict_top->address = (void *)addr;}
#define PARSELIT(name) {int lit; if (!getnumber(name, &lit)) goto error_undef; else push_ds(lit);}

    /* add primitives to host dictionary.
       they consist only of a code field, pointing to the goto labels below */

    /* return stack */
    PRIM("rclear", rdropall);
    PRIM("r", trs);
    PRIM(">r", tor);
    PRIM("r>", fromr);
    PRIM("rdrop", rdrop);
    PRIM("rsp@", getrsp);
    PRIM("rsp!", setrsp);

    /* control flow */
    PRIM(";_",  semi);
    PRIM("exit", exit);      xt_exit = last_xt;
    PRIM("enter", enter);
    PRIM("skip", skip);
    PRIM("jmp", jmp);        xt_jmp = last_xt;
    PRIM("donext", donext);
    PRIM("skipz", skipz);
    PRIM("jnz", jmpnz);
    PRIM("jz", jmpz);

    /* data stack */
    PRIM("cells", cells);
    PRIM("h@", hostfetch);
    PRIM("h!", hoststore);
    PRIM("usage", hostcells);
    PRIM("hc@", hostcfetch);
    PRIM("hc!", hostcstore);
    PRIM("dclear", dropall);
    PRIM("dup", dup);
    PRIM("drop", drop);
    PRIM("swap", swap);
    PRIM("over", over);
    PRIM("+",  add);
    PRIM("-",  sub);
    PRIM("*",  mul);
    PRIM("/",  div);
    PRIM("%",  mod);
    PRIM("and", and);
    PRIM("or",  or);
    PRIM("xor", xor);
    PRIM("<<", shl);
    PRIM(">>", shr);
    PRIM("<", lt);
    PRIM(">", gt);
    PRIM("<=", lte);
    PRIM(">=", gte);
    PRIM("=", eq);
    PRIM("random", random);
    PRIM("seed", seed);

    /* words & dict */
    PRIM("accept", accept);
    PRIM("buffer", buffer);
    PRIM("load", load);
    PRIM("(load)", _load);
    //PRIM("load#", load);
    PRIM("words", words);
    PRIM("macros", macros);
    PRIM("host", host);
    //PRIM("host#", host);
    PRIM("target", target);
    PRIM("macro", macro);
    PRIM("cell", cell);

    /* code field type primitives  */
    PRIM("doconst_", doconstbar);
    PRIM("lit", lit);                     xt_lit = last_xt;
    PRIM("dovar_", dovarbar);

    /* printing */
    PRIM("cr",  cr);
    PRIM("emit",  emit);
    PRIM(".tdict",  tdump);

    PRIM("in",  readport);
    PRIM("out", writeport);
    PRIM("read-fd",  readfd);
    PRIM("write-fd", writefd);
    PRIM("charmode",  charmode);


    PRIM("break", serialbreak);
    PRIM("iflush", serialiflush);
    PRIM("baud",   serialbaud);
    PRIM("sleep",  sleep);
    PRIM("usleep",  usleep);
    PRIM(".",  printdec);
    PRIM(".x",  printhex);
    PRIM(".S",  printdata);
    PRIM(".R",  printreturn);
    PRIM("output-file", outputfile);
    PRIM("(output-file)", _outputfile);
    PRIM("output-close", outputclose);
    

    /* interpretation + compilation */
    PRIM("accept",  accept);              xt_accept = last_xt;
    PRIM("(number)", _host_number);
    PRIM("interpret",  interpret);        xt_interpret = last_xt;
    PRIM(":",  colon);
    PRIM("(:)",  colon1);
    PRIM("enter_",  enterbar);
    PRIM("(hword)",  hword);
    PRIM("lastxt", lastxt);
    PRIM("here", here);
    PRIM("here!", sethere);
    PRIM("[,", interpretmode);
    PRIM("[_", interpretmode);
    PRIM("]",  compilemode);
    PRIM("execute", execute);
    PRIM("abort", abort);
    PRIM("quit", quit);    xt_quit = last_xt;
    PRIM("bye", bye);
    PRIM("hempty", empty);
    PRIM("hmark", mark);    xt_mark = last_xt; 
    PRIM("(", comment);
    PRIM("(,", comment);
    PRIM("(_", comment);
    PRIM("\\", lcomment);
    PRIM("\\,", lcomment);
    PRIM("\\_", lcomment);

    /* errors */
    PRIM("e-nolit", error_nolit);
    PRIM("e-page",  error_page);
    PRIM("e-undef", error_undef);
    PRIM("e-wisp",  error_wisp);

    /* host macros 
       since host is not the center of attention, host's version
       of ' postpone @ !   are prepended with h, and host , is _
     */
    PRIM("lit_", barlit);
    PRIM("_", bar);    xt_bar = last_xt;
    PRIM("char_", bar1);
    PRIM("align", align);
    PRIM("(hfind)", hfind);
    PRIM("(tfind)", tfind);
    PRIM("rfind", rfind);

    /* target code and dictionary */
    PRIM("(tdict)", tdict);
    PRIM("erecord", erecord);
    PRIM("attrib@", getattrib);
    PRIM("attrib!", setattrib);
    PRIM("unattrib", unattrib);

    /* target interpreter vectors */
    VARIABLE("<compile,>", 0);  xt_vcompilecomma = last_xt;   
    VARIABLE("<lit,>", 0);      xt_vlitcomma = last_xt; 
    VARIABLE("<texecute>", 0);  xt_vtexecute = last_xt;
    VARIABLE("<tpush>", 0);     xt_vtpush = last_xt; 


    // words();  print_wordlist();
    // macros(); print_wordlist();

    /* compile boot code */
    ip = (void ***)host_top; // we'll start here
    COMPILE(xt_mark);

    /* compile interpreter loop */
    interpreter_start = (void ***)host_top;
    COMPILE(xt_accept);
    COMPILE(xt_interpret);
    COMPILE(xt_jmp);
    COMPILE(interpreter_start);

    /* install c code error handler
       and jump to boot code */
    if (!sigsetjmp(jump_abort, 1)) NEXT;

    /* jmp_abort falls into abort */

    /******************** PRIMITIVES ********************/

    /* reset & exit to interpreter, 
       or exit program if not interactive */
  abort:
    while(!equal(top_is()->name, "-")) drop_is(); // this will exit if there's no stdin '-'
    stack_clear(ds);
    // current_target = HOST; // maybe this is a bad idea??
  quit:
    stack_clear(rs);
    compile_mode = 0;
    ip = interpreter_start;
    NEXT;
  bye:
    while(1) drop_is();
  mark:
    mark_dict_top  = dict_top;
    mark_host_top  = host_top;
    NEXT;
  empty:
    dict_top  = mark_dict_top;
    host_top  = mark_host_top;
    NEXT;
    
  align:
    align(); NEXT;

  /* parsing words : these are for bootstrapping text input.
     most of them (should) have their non-parsing equivalent working
     on the current word. */
  comment:
    do {nextword(); i=strlen(word); i=word[i-1];} 
    while (i!=')'); NEXT;
  lcomment:
    do {i = input_getc(); } while (i!='\n'); NEXT;
  colon:
    nextword(); goto colon1;
  accept:
    nextword(); NEXT;
  buffer:
    PUSH_DS(word); NEXT;
  load:     
    nextword(); push_is_file(word); NEXT;
  _load:
    push_is_file((char *)POP_DS()); NEXT;
  outputfile:        
    nextword(); push_os_file(word); NEXT;
  _outputfile:        
    push_os_file((char *)POP_DS()); NEXT;


  /* these are for bootstrapping compilation. 
     they can be redefined later. */
  colon1:
    WORD(word);          // add the word to the dictionary, pointing at the CW
    ADDR(host_top);      // register address
    COMPILE(&&enter);    // CW == enter
    compile_mode = 1;
    NEXT;
  semi:
    compile_mode = 0;
    COMPILE(xt_exit);
    NEXT;
  barlit:
    COMPILE(xt_lit);
  bar:
    i = pop_ds(); COMPILE(i); NEXT;
  bar1: 
    { 
	char *c = (char *)host_top;
	*c++ = pop_ds();
	host_top = (prim_t*)c;
    }
    NEXT;


#define FIND_BEGIN lookup_init(); while(1) { switch(lookup_tick(word)) {
#define FIND_END   default: lookup_continue(); } }
    
  /* compile and interpret */
  interpretmode:     
    compile_mode = 0; NEXT;
  compilemode:    
    compile_mode = 1; NEXT;

  /* dict manips */
  rfind:
    iword = rlookup((void*)POP_DS());
    if (iword) PUSH_DS(iword->name); else PUSH_DS(0);
    NEXT;

  tfind:
    lookup_init();
    while(1){
	switch(lookup_tick(word)){
	case MATCH(0,0,TARGET): PUSH_DS(iword->address); PUSH_DS(iword->bytes); NEXT;
	case -1: goto error_undef;
	default:
	    lookup_continue();
	}
    }
  hfind:
    lookup_init();
    while(1){
	switch(lookup_tick(word)){
	case MATCH(0,0,HOST): PUSH_DS(iword->address); PUSH_DS(iword->bytes); NEXT;
	case -1: goto error_undef;
	default:
	    lookup_continue();
	}
    }
  setattrib:    
    i = strlen(word); word[i] = POP_DS(); word[i+1] = 0; NEXT;
  getattrib:    
    i = strlen(word); if (i < 2) {PUSH_DS(0); NEXT; }
    switch(word[i-1]){
    case ',':
    case '_':
    case '#':
	PUSH_DS(word[i-1]);
    default:
	PUSH_DS(0);
    }
    NEXT;
  unattrib: 
    i = strlen(word); word[i-1] = 0; NEXT;
  
  hword:   WORD(word); ADDR(host_top); NEXT;
  tdict:   TARGETDICT(word, pop_ds()); NEXT;
  erecord: 
    iword = dict_top;
    iword->bytes = POP_DS() - (int)iword->address;
    // fprintf(stderr, "%d ", iword->bytes);
    NEXT;
  


    /* generic interpreter.
       basicly, this is a loop over the dictionary.
     */

  map_begin:
    lookup_init(); NEXT;
  map_continue:
    lookup_continue(); NEXT;
  map_tick:
    DS_PUSH(lookup_tick(word)); NEXT;
  map_match: 
    {   // ( in-attrib dict-attrib target -- packed ) 
	int t  = POP_DS(); // target
	int da = POP_DS(); // dict attrib
	int ia = POP_DS(); // input attrib
	DS_PUSH(MATCH(ia, da, t));
	NEXT;
    }
    
    
  


    /* compile mode 
       these are a bit of a hack because we don't have sanely encoded
       semantics. this is a thorn in my eye on the implementation side,
       but it's so easy and concise when using it i think i'm going to keep it.

       we use name mangling. matching is done in 2 stages:

       * find a possible match:
           names match, but one of the 2 words contains
           an extra character, which is treated as an attribute
	   pass this attribute as a match signature.

       * dispatch match signature 
           different combinations of input,dict name attributes and HOST/TARGET
           have a different interpretation. for each possible match, route
	   the different interpretations or try next match if no match.
     */

  enterbar:
    COMPILE(&&enter);
    //compile_mode = 1; ????
    NEXT;

    /* host mode (meta mode) */
  hcompile:
    lookup_init();
    while (1){
	switch(lookup_tick(word)){
	case MATCH(  0,   0, HOST):   goto _host_compile;    // compile host word
	case MATCH(  0, '_', HOST):   goto _host_execute;    // execute host macro
	case MATCH('_',   0, HOST):   goto _host_postpone;   // host postpone
	case MATCH('#',   0, TARGET): goto _host_compile_remote;   // compile remote call
	case MATCH(',',   0, TARGET): goto _target_postpone;  // target postpone
	case -1:                      goto _host_literal;
	default:
	    lookup_continue();
	}
    }

    /* target mode */
  tcompile:
    lookup_init();
    while (1){
	switch(lookup_tick(word)){
	case MATCH(  0,   0, TARGET):  goto _target_compile;      // compile target word
	case MATCH(  0, ',', HOST):    goto _host_execute;    // execute target macro
	case -1:                       goto _target_literal;
	default:
	    lookup_continue();
	}
    }

    /* target macro mode */
  mcompile:
    lookup_init();
    while (1){
	switch(lookup_tick(word)){
	case MATCH(  0,   0, TARGET):  goto _target_postpone;    // target postpone
	case MATCH(  0, '%', HOST):    goto _host_execute;      // execute macro immediate
	case MATCH(  0, ',', HOST):    goto _host_compile;        // compile target macro
	case -1:                       goto _target_postpone_literal; // target postpone literal
	default:
	    lookup_continue();
	}
    }

    /* host intepret mode */   
  hinterpret:
    lookup_init();
    while (1){
	switch(lookup_tick(word)){
	case MATCH(  0,   0, HOST):    goto _host_execute;  // host execute
	case MATCH('#',   0, TARGET):  goto _target_remote_execute; // target execute
	case -1:                       goto _host_number;    // host interactive number
	default:
	    lookup_continue();
	}
    }    

    /* target interpret */
  tinterpret:
    lookup_init();
    while (1){
	switch(lookup_tick(word)){
	case MATCH(  0,   0, HOST):  
	case MATCH(  0, '#', HOST):    goto _host_execute; // host execute
	case MATCH(  0,   0, TARGET):  goto _target_remote_execute; // target execute
	case -1:                       goto _host_number;   // host interactive number
	default:
	    lookup_continue();
	}
    }    

    /* macro interpret mode */
  minterpret:
    lookup_init();
    while (1){
	switch(lookup_tick(word)){
	case MATCH(  0,   0, HOST):  
	case MATCH(  0, '#', HOST):  
	case MATCH(  0, '%', HOST):  goto _host_execute; // host execute
	case -1:                     goto _host_number;   // host interactive number
	default:
	    lookup_continue();
	}
    }    

    
  _host_execute:
    w = iword->address; goto**w;
  _host_compile:
    w = iword->address; PUSH_DS(w); goto bar;
  _target_compile:
    w = iword->address; PUSH_DS(w); goto compilecomma;
  _host_literal:
    PARSELIT(word); goto barlit;
  _target_literal:
    PARSELIT(word); goto litcomma;
  _target_remote_execute:
    w = iword->address; PUSH_DS(w); goto texecute;
  tpush1:
    PARSELIT(word); goto tpush;

  _target_postpone:
    if (!iword->address) fprintf(stderr, "warning: %s is NULL\n", word);
    COMPILE(xt_lit);
    COMPILE(iword->address);
    COMPILE(((int**)xt_vcompilecomma)[1]);
    goto next;
  _target_postpone_literal:
    COMPILE(xt_lit);
    PARSELIT(word);
    COMPILE(POP_DS());
    COMPILE(((int**)xt_vlitcomma)[1]);
    goto next;
  _host_postpone:
    COMPILE(xt_lit);
    COMPILE(iword->address);
    COMPILE(xt_bar);
    goto next;

  _host_compile_remote:
    COMPILE(xt_lit);
    COMPILE(iword->address);
    COMPILE(((int**)xt_vtexecute)[1]);    
    goto next;    
  lnumber1:
    COMPILE(xt_lit);
    PARSELIT(word);
    COMPILE(POP_DS());
    COMPILE(((int**)xt_vtpush)[1]);    
    goto next;    
    
    
    

  _host_number:
    PARSELIT(word);
    NEXT;

    /* outer interpreter */
  interpret:
    if (compile_mode){
	switch(current_target){
	case HOST:   goto hcompile;
	case TARGET: goto tcompile;
	case MACRO:  goto mcompile;
	}
    }
    else {
	switch(current_target){
	case HOST:   goto hinterpret;
	case TARGET: goto tinterpret;
	case MACRO:  goto minterpret;
	}
    }
  execute:
    w = (void **)pop_ds();
    goto **w;
	  
  next:
    //iword = dict_rlookup(*ip);
    //printf("NEXT ip=0x%08x w=0x%08x *w=0x%08x %s\n", ip, *ip, **ip, iword ? iword->name : "unknown");
    w = *ip++;  // get pointer to code field of next word
    goto **w;   // jump to code pointed to by code field;
	
  /* return stack */
  rdropall: stack_clear(rs); NEXT;
  trs:      PUSH_DS(top_rs()); NEXT;
  tor:      PUSH_RS(POP_DS()); NEXT;
  rdrop:    POP_RS(); NEXT;
  fromr:    PUSH_DS(POP_RS()); NEXT;
  setrsp:   stack_set_rsp(rs, pop_ds()); NEXT;
  getrsp:   PUSH_DS(stack_get_rsp(rs)); NEXT;

  /* control flow */
  exit:     ip = POP_RS(); NEXT;
  enter:    PUSH_RS(ip); ip = (void ***)(w+1); NEXT;
  jmp:      ip = (void ***)(ip[0]); NEXT;
  skip:     ip++; NEXT;
  donext:   {int *pi = (int *)(rs->top); if (--(pi[0])) goto jmp; goto skip;}
  skipz:    if (!POP_DS()) goto skip; NEXT;
  jmpnz:    if (POP_DS()) goto jmp; goto skip;
  jmpz:     if (!POP_DS())  goto jmp; goto skip;
  

  /* data stack */
  cells:   {int i = pop_ds(); i *= sizeof(int); push_ds(i); NEXT; }

  hostfetch:   {int *pi = (int *)pop_ds(); push_ds(pi[0]); NEXT; }
  hoststore:   {int *pi = (int *)pop_ds(); pi[0] = pop_ds(); NEXT; }
  hostcfetch:  {char *pc = (char *)pop_ds(); push_ds(pc[0] & 0xff); NEXT; } // UNSIGNED!!
  hostcstore:  {char *pc = (char *)pop_ds(); pc[0] = pop_ds(); NEXT; }
  hostcells:   {push_ds(host_top - host_buffer); NEXT; }

  dropall: stack_clear(ds); NEXT;
  dup:     PUSH_DS(top_ds()); NEXT;
  drop:    POP_DS(); NEXT;
  swap:    {int i = POP_DS(); int j = POP_DS(); PUSH_DS(i); PUSH_DS(j); NEXT;}
  over:    {int *pi = (int *)(ds->top); PUSH_DS(pi[1]); NEXT;}

  /* math */
  add:    OP2(+);
  sub:    OP2(-);
  mul:    OP2(*);
  div:    OP2(/);
  mod:    OP2(%);
  and:    OP2(&);
  or:     OP2(|);
  xor:    OP2(^);
  shl:    OP2(<<);
  shr:    OP2(>>);
  lt:     CMP(<);
  gt:     CMP(>);
  eq:     CMP(==);
  lte:    CMP(<=);
  gte:    CMP(>=);

  random: push_ds(random()); NEXT;
  seed:   srandom(pop_ds()); NEXT;
  

  /* io */

  /* misc */
  words:    words(); print_wordlist(); NEXT;
  macros:   macros(); print_wordlist(); NEXT;
  host:     current_target = HOST;   NEXT;
  target:   current_target = TARGET; NEXT;
  macro:    current_target = MACRO;  NEXT;
  lastxt:   PUSH_DS(last_xt); NEXT;
  here:     PUSH_DS(host_top); NEXT;
  sethere:  host_top = (prim_t *)POP_DS(); NEXT;
  cell:     PUSH_DS(POP_DS() * sizeof(int)); NEXT;

  /* do things */
  doconst:     PUSH_DS(((int *)w)[1]); NEXT;     // number stored in dictionary
  doconstbar:  COMPILE(&&doconst); NEXT;
  lit:         PUSH_DS(((int)(*ip++))); NEXT;    // number stored in code
  dovar:       PUSH_DS(w+1); NEXT;
  dovarbar:    COMPILE(&&dovar); NEXT;

  /* port io */
  writefd:   
  { 
      int fd = POP_DS();
      char buf[1]; 
      buf[0] = pop_ds(); 
      write(fd, buf, 1); 
      NEXT; 
  }
  readfd:    
  { 
      int fd = POP_DS();
      char buf[1]; 
      if (1 != read(fd, buf, 1)) { 
	  perror(""); 
	  error_abort("can't read"); 
      }
      push_ds(buf[0]); 
      NEXT; 
  }
  readport:    PUSH_DS(port_fd); goto readfd;
  writeport:   PUSH_DS(port_fd); goto writefd;


  sleep:       sleep(pop_ds()); NEXT;
  usleep:      usleep(pop_ds()); NEXT;
  serialbreak:  tcsendbreak(port_fd, 0); NEXT;
  serialiflush: tcflush(port_fd, TCIFLUSH); NEXT;
  serialbaud: {
      int baud = POP_DS();
      switch(baud){
      default:
      case 9600:   baud = B9600;  break;
      case 19200:  baud = B19200; break;
      case 38400:  baud = B38400; break;
      case 57600:  baud = B57600; break;
      case 115200: baud = B115200; break;
	  error_abort("invalid baud rate");
      }
      struct termios t;
      if (-1 == tcgetattr(port_fd, &t)) error_abort("can't get terminal attributes");
      if (-1 == cfsetospeed(&t, baud))  error_abort("can't set output baud rate");
      if (-1 == cfsetispeed(&t, baud))  error_abort("can't set input baud rate");
      if (-1 == tcsetattr(port_fd, TCSANOW, &t)) error_abort("can't set terminal attributes");
      NEXT;
  }

  charmode:
    {
	static struct termios stored_settings;
	if (POP_DS()){
	    struct termios new_settings;
	    tcgetattr(0,&stored_settings);
	    new_settings = stored_settings;

	    /* Disable canonical mode, and set buffer size to 1 byte */
	    new_settings.c_lflag &= (~ICANON);
	    new_settings.c_lflag &= (~ECHO);
	    new_settings.c_cc[VTIME] = 0;
	    new_settings.c_cc[VMIN] = 1;

	    tcsetattr(0,TCSANOW,&new_settings);
	    NEXT;
	}
	else {
	    tcsetattr(0,TCSANOW,&stored_settings);
	    NEXT;
	}
    }


  /* printing */
  cr:          output_printf("\n"); NEXT;
  printdec:    output_printf(" %d", pop_ds()); NEXT;
  printhex:    output_printf(" 0x%08x", pop_ds()); NEXT;
  printdata:   stack_print("%d ", ds); NEXT;
  printreturn: stack_print("0x%08x ", rs); NEXT;
  emit:        output_printf("%c", pop_ds()); NEXT;

  /* dump target dict */
  tdump: 
  {
      item_t **tmp = (item_t **)host_top; // temp buf for reversing
      int count = 0;
      iword = dict_top;
      while (iword != 0){
	  if (iword->target == TARGET){ 
	      //fprintf(stderr, "adding %s\n", iword->name);
	      tmp[count++] = iword; 
	  }
	  iword = iword->prev;
      }
      while (count--){
	  output_printf("0x%04x 0x%04x entry %s\n", 
			tmp[count]->address, 
			tmp[count]->address+tmp[count]->bytes, 
			tmp[count]->name);
      }
      NEXT;
  }


  /* errors */
  error_wisp:   error_abort("wisp programmer error.");
  error_nolit:  error_abort("construct not implemented for non-literal arguments.");
  error_page:   error_abort("page boundary error.");
  error_undef:  error_undef(word); goto abort;


  /* files */
  outputclose:       drop_os(); NEXT;

  /* target compilation hooks */
  litcomma:     w = xt_vlitcomma;     goto w1execute;
  compilecomma: w = xt_vcompilecomma; goto w1execute;
  texecute:     w = xt_vtexecute;     goto w1execute;
  tpush:        w = xt_vtpush;        goto w1execute;
  w1execute:    w=w[1]; goto **w;

}


