( -*- forth -*- )

( CrapSynth 29  --- [c] 2005 Tom Schouten

This program is free for non-commercial use.
Contact the author for more information. )



( simple sound synthesis engine - runs at 8 khz.
this leaves us 125 cycles per sample to do our magic.

osc0: main [pitched] oscillator
osc1: modulator 1
osc2: modulator 2

)

host
0x0b0 constant kernel ( space before reserved for xt table )
10    constant calib  ( osc calibration )


load stc14/12f675.f   ( forth macros )
load stc14/forth.f    ( forth lib )
load stc14/eeprom.f   ( eeprom macros )

target
0x1 constant xt-forth
0x2 constant xt-initsynth
0x3 constant xt-sequencer




( SYNTH PARAMETERS )

( oscillator periods )
variable f-osc0
variable f-osc1
variable f-osc2

( sound effects )
variable chirp+     ( chirp increment )
variable range      ( chirp range )
variable octave     ( octave modifier for osc loading )

( current engine: combination of synth/control config )
variable engine     ( [ osc2 | control | osc1 | synth ] )

( oscillator enable flags )
engine 3 2constant osc1enable
engine 7 2constant osc2enable




( INTERNAL STATE VARS )

( state saving for receive and synth tasks
a task's data struct contains WREG and STATUS of
the interrupted context )

task t-synth
task t-receive



( frequency division counters )
variable p-cont    ( sample -> control freq )
variable p-note    ( control -> note freq )

variable p-trans   ( nb of control ticks to run transient synth )
variable p-delay   ( nb of notes to delay VM program )

variable p-osc0    ( osc0 phase. main oscillator )
variable p-osc1    ( osc1 phase, main modulator )
variable p-osc2    ( osc2 phase, aux modulator )

( noise generator state )
variable noise0
variable noise1

( interpreter state )
variable rsp     ( return stack )
variable esp     ( engine stack )
variable tsp     ( transient time stack )
variable ip      ( instruction pointer )
variable ins     ( current instruction )

( receiver state )
variable r-state   ( receiver state )
variable p-bit     ( bit counter )
variable shiftreg  ( receiver shift register )
variable data      ( received data )


( synth output buffer )
variable output    ( copied to output in synth isr )

( event flags )
flag  icont      ( controller tick )
flag  idata      ( receiver data ready )
flag  ibye       ( leave forth interpreter )

host

( this is the instruction rate )
1000000 constant cpufreq
8000    constant samplerate

( frequency division factors for
hierarchical time division )
40  neg constant CONT    ( 200hz )
25  neg constant NOTE    ( 1/4 beat @ 120 bpm )


( convert frequency to oscillator period: freq -- period )
: hz  samplerate 2 / swap / neg ;
: ms  samplerate CONT / samplerate 2 / swap / neg ;

load synth/host.f ( host defs )

macro

( COUNTERS )
: cont?    p-cont  1+! z? ;
: note?    p-note  1+! z? ;
: delay?   p-delay 1+! z? ;
: trans?   p-trans 1+! z? ;

: osc0?    p-osc0  1+! z? ;
: osc1?    p-osc1  1+! z? ;
: osc2?    p-osc2  1+! z? ;

: bit?     p-bit   1+! z? ;


( INIT )


macro

: 0note [ NOTE ]L p-note ! ;
: 0cont [ CONT ]L p-cont ! ;
: 0all  0note 0cont ;




macro

: softint cont? if 0cont icont high then ; ( controller freq dev )


( schedule next timer tick )
: timer
    INTCON T0IF low      ( clear interrupt bit )
    [ cpufreq samplerate ( compute period )
    / neg 2 +            ( adjust for 2 cycle non-increment after writing )
    calib + ]L           ( manual calibration )
    TMR0 +! ;            ( add to timer value to compensate for time spent )

( IOC handling )
: ioc
    GPIO @ drop 
    INTCON GPIF low ;    ( clear interrupt bit )

: int-enable
    INTCON GPIE high    ( enable GPIO interrupt )
    INTCON T0IE high ;  ( enable timer interrupt )


: ioc?   INTCON GPIF high? ;
: timer? INTCON T0IF high? ;

( OSCILLATOR )

( pin status )
macro
( : output      GPIO ; )        
: audio       output 5 ;
: led         output 2 ;
: toggle      0x20 output xor! ;

: on?     audio high? ;
: off?    audio low? ;
: on      audio high ;
: off     audio low ;

: input   GPIO 4 ;
: data?   input low? ; ( falling edge is pwm data edge 1/3 or 2/3 )
target


( synth done, restore saved state and iret )
: synth-done int-enable t-synth context@ [ RETFIE , ] ;  undo



( osc2 modulator modulates the output
of the other synths if it's on  )
: mod2 osc2enable high? if osc2? if f-osc2 p-osc2 @! toggle then then synth-done ;

( osc0/osc1 independent, cross modulation
can be used to generate pwm too )
: xmod
    ( 3 independent oscillators, output is multiplied )
    osc0? if f-osc0 p-osc0 @! toggle then 
    osc1enable high? if osc1? if f-osc1 p-osc1 @! toggle then then
    mod2 ;

( osc0 turns on and resets osc0/osc1, osc1 turns on/off
can be used to generate virtual polyphony )
: sync
    osc0? if
	f-osc0 p-osc0 @!  ( reset main osc )
	f-osc1 p-osc1 @!  ( reset mod osc )
	on synth-done exit then
    osc1? if f-osc1 p-osc1 @! toggle then
    mod2 ;

( noise generator )
: noise [
    ( out = 15 x 14 x 12 x 3 )
    0 movlw,
    noise1 7 ( 15 ) btfsc, 1 xorlw,
    noise1 6 ( 14 ) btfsc, 1 xorlw,
    noise1 4 ( 12 ) btfsc, 1 xorlw,  
    noise0 3 ( 03 ) btfsc, 1 xorlw,

    ( put bit in carry bit )
    STATUS xorwf, ( this affects status )
    STATUS movwf, ( this won't )
    
    ( shift carry into shift reg )
    noise0 rlf, f,
    noise1 rlf, f,
    ]
    ( carry out will determine toggle )
    c? if toggle then synth-done ;

( similar, but modulated up/down by osc1
and passed through mod2. osc1 enable bit determines
up/down modulation )
: noisemod
    osc1enable high? if toggle then ( modulate up )
    osc1? if
	f-osc1 p-osc1 @!
	noise exit then
    mod2 ;

( bit pattern / wavetable synth
this is the main debugging output
osc0 : subdivision
osc1 : bit counter
osc2 : f-osc2 contains pattern )
: pattern
    osc0? if
	f-osc0 p-osc0 @!                ( reset subdivision counter )
	osc1? if
	    f-osc1 p-osc1 @!            ( reset bit counter )
	    osc2enable high? if
		f-osc2 p-osc2 @!        ( bit pattern from f-osc2 )
	    else
		data p-osc2 @!          ( bit pattern from received data buffer )
	    then
	    led high                    ( + scope sync pulse )
	else
	    led low                     ( - scope sync pulse )
	then
	[ clc, p-osc2 rlf, f, ]         ( msb first : easier to read on scope )
	c? if on then nc? if off then   ( c -> pin )
    then synth-done ;

: none synth-done ;

( output last received bit )
: received
    off shiftreg 0 high? if on then synth-done ;
    
    


( jump to current synth engine )
: synth engine @ 0x7 and route
    xmod pattern sync noise
    noisemod received none none ;


( RECEIVER )

( period = 125 cycles  [make this variable later]
1 = [110] 2/3 long cycle  = 83 cycles
0 = [100] 1/3 short cycle = 41 cycles
so we check at 62 )

macro
: c>led c? if led high else led low then ;
: stopwatch 0 TMR0 ! ; ( start counting )

host
: bit>c, 
    -70 movlw, TMR0 addwf, ( c = TMR0 > 70 )
    ( f-osc2# movwf, ) ;   ( debug )

( the timer gets set 13 cycles after handling
the up edge isr. it's read about 10 cycles later
relatively, so we compare at 70.

after measuring i find
bit    TMR0 70 -
1      +12/13
0      -10

this seems perfect actually. so all errors should
be protocol or overflow errors. )

target

macro ( receiver state switching )
: IGNORE  0 r-state ! ;
: IDLE    1 r-state ! ;
: RECV    2 r-state ! -8 p-bit ! ;
: STOP    3 r-state ! ;

target

( forth interpreter )
: read        begin idata high? until idata low data @ ;
: execute     PCL ! ; undo

macro
( check if there's a command, if so, execute )
: poll idata high? if read execute then ;
target

( receive done, restore saved state and iret )
: receive-done int-enable t-receive context@ [ RETFIE , ] ;  undo


( states: each state reads the timer
at the start, to minimise jitter. )

( receiver off: when switching from timer to ioc interrupt,
we ignore the first falling edge )
: ignore
    IDLE receive-done ;

( ready for start bit )
: idle
    bit>c nc? if RECV then receive-done ;

( ready for data bit )
: receive
    bit>c shiftreg c>>!      ( store bit in shift reg )
    bit? if                  ( count bit )
	STOP                 ( set next state if 8 bits )
    then receive-done ;

( ready for stop bit )
( stop bit is not really necessary,
but makes it easier to switch to other layer,
+ ignores data from a 100 pwm signal [110 is idle] )
: stop
    bit>c c? if              ( if valid stop bit )
	shiftreg data @!     ( store received byte )
	idata high           ( set software interrupt )
    then IDLE receive-done ; ( next state is IDLE )

( receiver state routing )
: comm
    r-state @ 3 and route
    ignore idle receive stop ;



( all isr routines call 'done' at the end to set the timer
execution: uses 1 return stack cell, does not use FSR )

(
timer mode -> isr saves to synth context, synth run with ints disabled
ioc mode ->
+ saves to synth context, synth is run with ints enabled
- saves to receive context

probably for ioc need to save to receive context first,
then if it's a positive edge, move receive->synth context.
-> can only check GPIO with bank set to 0

)


macro
: out output GPIO @! ; 
target

there 0x6 xt!
: ioc-isr
    t-receive context! bank0 ioc
    data? if comm exit then      ( data edge -> receiver )
    stopwatch                    ( mark edge )
    out                          ( copy buffer to port here to reduce jitter )
    t-receive t-synth context@!  ( move context )
    sti                          ( end isr )
    softint                      ( perform softint frequency division )
    synth                        ( spawn synth task )
; 

there 0x5 xt!
: timer-isr 
    t-synth context! bank0 out
    IGNORE timer softint synth ;

( dispatch interrupt. compiles
btfsc INTCON, 2
at the isr location, which chooses ioc-isr
or timer-isr depending on ioc bit.
see disassembly 0 .d )

( 0x190b 0x8 t! )
there 0x8 there!
    INTCON T0IF btfsc,
there!




( EVENT HANDLERS )

( simple note sequencer

we got 128 bytes of eeprom. this could be used
as sequencer pattern data, or bytecode for a simple
interpreter running at note rate. we go for the latter
one because it's cooler

infrastructure:
* instruction pointer
* 2 stack pointers
* stacks: return, engine


what about a single stack for synth/control config?
a lot easier to implement, and it should solve synchro
problems between the 2, mainly for transients.

got 8 bits:
[ mod1 | mod2 | synth | control ]

TODO:
* smaller stack
* movf jumptable ipv indirect @
* small buffer for comm bytes




)


host
: +sp, <lit, incf, f, ; ( -- )
: -sp, <lit, decf, f, ; ( -- )

macro
: rs  rsp @ 0x7 and 0x50 + ;
: es  esp @ 0x3 and 0x58 + ;
: ts  tsp @ 0x3 and 0x5c + ;

target

( VM return stack )
: >rs rsp +sp rs ! ;
: rs> rs @ rsp -sp ;

( synth/control stack )
: >es esp +sp es ! ;
: es> es @ esp -sp ;

( transient time stack )
: >ts tsp +sp ts ! ;
: ts> ts @ tsp -sp ;

( set/push ops: x -- )
: period0!    not f-osc0 ! ;
: period1!    not f-osc1 ! ;
: period2!    not f-osc2 ! ;
: >engine     engine @ >es engine ! ;
: engine!     engine ! ;
: >trans      p-trans @ >ts not p-trans ! ;


( lookup tables for note scale
equal tempering, 16 notes C -> D#

low C is at 32.704 hz,
low A is at 55.000 hz )

: scale 0xf and table
    -246 -232 -219 -207
    -195 -184 -174 -164
    -155 -146 -138 -131
    -123 -116 -110 -104 ;

( counter scale -- counter )
: transpose [
    INDF movf, drop,
    0 xorlw, ( for flags )
    there
    STATUS Z btfsc, exit,
    stc, INDF rrf, f,
    0xff addlw,
    jump,
    ] ; undo

: frequency scale octave @ transpose ;

: transient not ;


( VM INSTRUCTIONS + DECODING )


: (exit) -1 p-delay ! ;  ( undefined opcode: exit interpreter )

: ee@ (ee@) ;

( advance interpreter. set ins and next ip )
: advance 
    ip @ ee@ ins !
    ip 1+! 0x7f ip and! ;


( if used once, make it a macro
also, route optimizes some push/pop )
macro 
: 1:0 ins @ 0x3 and ;
: 5:4 ins @ swapdigits 0x3 and ;
: 4:3 ins @ r<<c swapdigits 0x3 and ;
: 2:0 ins @ 0x7 and ;
: 3:0 ins @ 0xf and ;
: 3:1 ins @ c>>r 0x7 and ;
: 4:1 ins @ c>>r 0xf and ;

target

( 0x00  [ 0 | aaaaaaa ] call subroutine )
: subroutine ip @ >rs ins ip @! ;


( [ 10 | xx | dddd ] 4 bit literals
0x80 f-osc0
0x90 f-osc1
0xa0 f-osc2
0xb0 delay )

( x -- )
: note0! frequency f-osc0 ! ;  
: note1! frequency f-osc1 ! ;  
: note2! frequency f-osc2 ! ;  
: delay! not p-delay ! ;

: 4lit 3:0 5:4 route
    note0! note1! note2! delay! ;

( [ 110 | xx | ddd ] 3 bit literals
0xC0 noise burst [hihat/snare]
0xC8 octave select
0xD0 frequency inc/dec [chirp]
0xD8 frequency range [chirp] )

: (drop) drop ;
: chirp!  not chirp+ ! ;
: octave! octave ! ;
: range!  0 swap transpose range ! ;
: burst!
    >trans ( set transient timing )
    [ 0 3 0 0 pack-engine ]L >engine ;

: 3lit 2:0 4:3 route
    burst! octave! chirp! range! ;

( 0xE0 [ 1110 | xxx | d ] 1 bit literals
special case 1 bit literals )
: 1lit 4:1 route
    (exit) (exit) (exit) (exit)
    (exit) (exit) (exit) (exit) ;

( 0xF0 [ 11110 | xxx ] 8 bit operand commands 
0xF0 period0!
0xF1 period1!
0xF2 period2!
0xF3 >engine
0xF4 engine!
0xF5 >trans
0xF6
0xF7 )

: 0lit 2:0 advance ins @ swap route
    period0! period1! period2! >engine
    engine! >trans (exit) (exit) ;
    

( [ 111110 | xx ] pop instructions
0xF8 ret
0xF9 drop-engine
0xFA dup-engine
0xFB swap-engine )

: rs>ip        rs> ip ! ;
: drop-engine  es> engine ! ;
: dup-engine   engine @ >es ;
: swap-engine  es> es> swap >es >es ;
: pop   1:0 route
    rs>ip drop-engine dup-engine swap-engine ;


( 0xFC [ 111111 | xx ] misc
0xFC pwm
0xFD
0xFE
0xFF nop -> empty eeprom )
: pwm
    f-osc0 f-osc1 @! 1 f-osc1 -!
    f-osc1 f-osc2 @! 1 f-osc2 -! ;
    
: misc 1:0 route
    pwm (exit) (exit) (exit) ;

( toplevel decoder )
: interpret
    ins 7 low? if subroutine exit then
    ins 6 low? if 4lit exit then 
    ins 5 low? if 3lit exit then
    ins 4 low? if 1lit exit then
    ins 3 low? if 0lit exit then
    ins 2 low? if pop exit then
    misc ;

( VM interpreter
runs until p-delay is nonzero )
: interpreter
    begin
	poll       ( in case program gets stuck, we still have access to forth )
	advance    ( read instruction and increment instruction pointer )
	interpret  ( interpret instruction )
    p-delay test nz? until ;


( SOUND EFFECTS - control rate engine )



( control rate )

: (nothing)    ;
: chirp     chirp+ @ f-osc0 +! range @ f-osc0 or! ;
: woot      chirp+ @ f-osc0 -! range @ f-osc0 or! ;

( transient controller )
: trans
    0 p-trans xor! nz? if             ( check if in transient )
	trans? if                     ( yes, count and pop if done )
	    drop-engine               ( discard engine )
	    ts> p-trans !             ( restore prev transient time )
	then
    then ;

: control
    trans                               ( run transient counter )
    engine @ swapdigits 0x7 and route   ( dispatch controller engine )
    (nothing) chirp woot (nothing)
    (nothing) (nothing) (nothing) (nothing)
;


( MAIN LOOP )
macro

: digital      bank0 0 GPIO ! 7 CMCON ! bank1 0 ANSEL ! bank0 ;
: interrupts   INTCON PEIE high INTCON GIE high ;
: init-timer   bank1 clrwdt 0x88 OPTION_REG ! bank0 ;     ( pullup off, prescaler off )
: output-2/5   bank1 0x1b TRISIO ! bank0 ;                ( GP2/5 = out, rest = in )
: init-ioc     bank1 0x10 IOC ! bank0 ;                   ( GP4 interrupt on change )

target

( misc forth tokens defined after all the other xts
to prevent the table from shifting too much )

: bye ibye high ;   ( can't manip RS, so we use an exit flag. )
: ee! (ee!) IGNORE sti ;  ( ignore next comm bit )
: >ee ip @ ee! ip 1+! ;

( eeprom programming )
: program ( n start -- )
    ip ! begin read >ee 1- z? until drop ;

' read    xt
' (@)     xt
' (!)     xt
' bye     xt
' ee@     xt
' ee!     xt
' program xt

there xt-initsynth xt!
: initsynth
    ( init synth engine )
    -1 chirp+ !
    dup f-osc0 !
    dup f-osc1 !
    dup f-osc2 !
    dup p-delay !
    3 octave !
    0x55 noise0 !
    0x55 noise1 !
    0 engine !
;

( FORTH: executes only forth code,
to make sure there are no buffer overruns due to
execution of controller or interpreter )


there xt-forth xt!
: forth
    begin
	read execute
    ibye high? until
    ibye low ;

( SEQUENCER: executes code from eeprom )
there xt-sequencer xt!
: sequencer
    ( init interpreter )
    0x0 rsp !
    0x0 esp !
    0x0 ip !

    ( dispatch sofware interrupt )
    begin
	( data ready -> execute command )
	poll

	( controller timer expired -> execute control )
	icont high? if
	    icont low control

	    ( note + delay timer exipired -> execute interpreter )
	    note? if
		0note delay? if interpreter then
	    then
	then
    again ;


application : main
    cli                 ( in case of soft reset -> disable interrupts )
    0x60 FSR !          ( set data stack pointer to top of mem before clearmem )
    clearmem 0x50 FSR ! ( set data stack pointer below synth VM stacks )
    digital output-2/5
    0all init-timer init-ioc interrupts

    int-enable initsynth sequencer ;







(
todo:
* glide
* bassdrum opcode [special case of glide]
* better hihat/snare algos

alternative: on each [external] note sync:

* reset transients
* run note bytecode interpreter
* run controller code until external sync

* something with clicks.

transients with count on RS
this is possible, as long there's no function call after
the transient loading. maybe best to use a separate stack for this.


)


( external control commands:

commands that reset note interpreter:

* reprogram eeprom
* sync internal synth
  beat sync is not necessary, since we have sample sync.
  -> start/stop
  -> reset ip
  -> change note/control divisors

commands that interrupt note interpreter:

* state variable change

implementation is a forth VM. received bytes are execution tokens.
[code vectors from the first 256 word block]

)

( priorities.
i suspect a problem with detecting the falling edge of the clock,
when the synth is doing too much work:

worst case path for synth : 57 > 41, so we have a problem

so the synth is run as a pre-emptable task by saving the
interrupted context to t-synth instead of t-receive.
same goes for the forth interpreter, which is spawned
from the receive isr if not already running.

TIMER / GPIO+ falls into synth task
GPIO- executes receive isr

so:
timer? -> timer isr -> no problem here, since no receive
          so run everything in isr
gpio?  ->
     + -> spawn synth task
     - -> handle receive in isr


)

